#!/usr/bin/perl
#
# Encoding data from DVK MX DSK file to HxC hfe
#
# Supported DSK types:
# 40 tracks, 1 side , Sector 256 bytes, 11 sectors per track, 112640 bytes
# 40 tracks, 2 sides, Sector 256 bytes, 11 sectors per track, 225280 bytes
# 80 tracks, 1 side , Sector 256 bytes, 11 sectors per track, 225280 bytes
# 80 tracks, 2 sides, Sector 256 bytes, 11 sectors per track, 450560 bytes
#
# (c) 2104 PortaOne, Inc.
# Oleksandr Kapitanenko kapitan@portaone.com
#
# Usage:
# dvk-dsk-to-mx-hfe.pl <dsk file> <hfe file>
#
use POSIX qw(ceil);
$| = 1;

# Format version Old/New driver
#
my $MX_Driver = 'NEW';
# my $MX_Driver = 'OLD';

# Emulate "first word" artifact
#
my $First_Word_Artifact = 0;

# Debug flag
#
#my $debug = 'FM';

die "Output and input files are the same!" if $ARGV[0] eq $ARGV[1];
unlink($ARGV[1]);
open(IN, "<$ARGV[0]")
   || die "File $ARGV[0] not found.";
open(OUT, ">$ARGV[1]")
   || die "Can not create output file $ARGV[1]";

# Read DSK file
#
my $DSK_Tracks, $DSK_Sides;
my $dsk_size = -s $ARGV[0];
if( $dsk_size == 112640 ) {
    print "DSK 40 tracks, 1 side.\n";
    $DSK_Tracks = 40;
    $DSK_Sides = 1;
}
elsif( $dsk_size == 225280 ) {
    print "DSK 40 tracks, 2 sides OR 80 tracks, 1 side.\n";
    print "Assuming 40 tracks, 2 sides.\n";
    $DSK_Tracks = 40;
    $DSK_Sides = 2;
}
elsif( $dsk_size == 450560 ) {
    print "DSK 80 tracks, 2 sides.\n";
    $DSK_Tracks = 80;
    $DSK_Sides = 2;
} else {
    die "Not DSK file?";
}

print "Reading DSK  ";
my @dsk, $buff, $dsk_track, $track;
for( $track = 0; $track < $DSK_Tracks*2; $track++ ) {
   $dsk_track = [];
   if( $DSK_Sides == 1 && ($track & 0x01) == 1 ) {
      push $dsk_track, ( 0x00 ) x 256*11;
   } else {
      sysread(IN, $buff, 256*11 ); # Track 256 bytes * 11 sectors
      push $dsk_track, unpack( 'C2816', $buff);
   }
   $dsk[$track] = $dsk_track;
   print ".";
}

close(IN);

# DSK to TRK
#
print "\nDSK to TRK ";
my $data_word, $MSB, $LSB, $crc, $dsk_byte;

for( $track = 0; $track < $DSK_Tracks*2; $track++ ) {
   $dsk_track = [];
   $dsk_byte = 0;

   # "First Word" artefact
   #
   if( $First_Word_Artifact ) {
      if( $MX_Driver eq 'OLD' ) { 
         push $dsk_track, ( 0xFF, 0xFF ); 
      } else {
         $LSB = $track - 1;
         $LSB = 0 if $LSB < 0;

         push $dsk_track, ( 0x83, $LSB );
      }
   }

   # GAP1
   #
   if( $MX_Driver eq 'OLD' ) {
      push $dsk_track, ( 0x00 ) x 60;
   } else { 
      push $dsk_track, ( 0x00 ) x 16;
   }

   # Marker
   #
   push $dsk_track, ( 0x00, 0xF3);

   # Track Number
   #
   push $dsk_track, ( 0x00, int($track/2) );

   for($sector = 0; $sector < 11; $sector++) {

      # DATA
      #
      $crc = 0;

      for ($count = 128; $count >= 1; $count--) {
         $LSB = $dsk[$track][$dsk_byte++];
         $MSB = $dsk[$track][$dsk_byte++];
         $data_word = $MSB*256 + $LSB;

         $crc = ( $crc + $data_word ) & 0xFFFF;

         push $dsk_track, ( $MSB, $LSB );
      }

      # CRC
      #
      $MSB = $crc >> 8;
      $LSB = $crc & 0xFF;
      push $dsk_track, ( $MSB, $LSB );
   }

   if( $MX_Driver eq 'OLD' ) {
      push $dsk_track, ( 0x83, 0x01 ) x 2;
   } else {
      # Optional Track/Side information for new MX driver
      #
      $LSB = $track - 1;
      $LSB = 0 if $LSB < 0;

      push $dsk_track, ( 0x83, $LSB ) x 3;
   }

   # Total of 2864. Total FM track is 3125 bytes.
   # Remaides is 
   # New 261 or 259 in case of "first word artifact"
   # Old 249 or 247 in case of "first word artifact"
   #
   if( $MX_Driver eq 'OLD' ) {
      if( $First_Word_Artifact ) {
         push $dsk_track, ( 0xFF ) x 247;
      } else {
         push $dsk_track, ( 0xFF ) x 249;
      }
   } else {
      if( $First_Word_Artifact ) {
         push $dsk_track, ( 0x00 ) x 259;
      } else {
         push $dsk_track, ( 0x00 ) x 261;
      }
   }

   $dsk[$track] = $dsk_track;
   print ".";
}

# Generate FM code
#
print "\nEncoding FM ";

$fm_code[0] = 0x20;
$fm_code[1] = 0xA0;
 
my $mfm_track, $byte, $bit, $bit_count, $fm_i;
for( $track = 0;  $track < $DSK_Tracks*2; $track++ ) {
   $fm_i =0;

   for( $i = 0;  $i < $#{$dsk[$track]}+1; $i++ ) {
      $byte = $dsk[$track][$i];

      $bit_count = 0;
      $mfm_track[$track][$fm_i] = 0;
      while( $bit_count++ < 8 ) {
         $bit = ($byte & 0x80) >> 7;
         $byte = $byte << 1;
         if( $bit_count == 3 || $bit_count == 5 || $bit_count == 7 ) {
            
            # Advance to next FM byte
            #
            $fm_i++;
            $mfm_track[$track][$fm_i] = 0;
         } else {
            $mfm_track[$track][$fm_i] = $mfm_track[$track][$fm_i] >> 4;
         }
         $mfm_track[$track][$fm_i] = $mfm_track[$track][$fm_i] | $fm_code[$bit];
      }
      $fm_i++;
   }
   print ".";
}

print "\nCreating HFE ";

# Create HxC header
#
print OUT "HXCPICFE";
print OUT chr(0);           # Revision 0
print OUT chr($DSK_Tracks); # Number of tracks
print OUT chr($DSK_Sides);  # Number of sides (Not used by the emulator)
print OUT chr(2);           # Track Encoding mode FM
print OUT chr(250), chr(0); # Bitrate in Kbit/s. Ex : 250=250000bits/s for FM true_bitrate*2
print OUT chr(0), chr(0);   # Rotation per minute (Not used by the emulator)
print OUT chr(7);           # Floppy interface mode
print OUT chr(0);           # Write protected
print OUT chr(1), chr(0);   # Offset of the track list LUT in block of 512 bytes (Ex: 1=0x200)

# Advance to 0x200
#
print OUT chr(0xFF)x(0x200-20);

# Create track list
#
my $dsk_track_len = $#{$dsk[0]}+1;
my $offset = 2; # Start offset 2x512 bytes blocks
for( $track = 0; $track < $DSK_Tracks; $track++ ) {
   $track_len = $dsk_track_len * 4 *2;
   print OUT chr($offset&0x00ff), chr($offset>>8), chr($track_len&0x00ff), chr($track_len>>8);
   $offset += ceil($track_len/512);
}

# Advance to 0x400
#
print OUT chr(0xFF)x(0x200-$DSK_Tracks*4); 

# Dump FM code to hfe file
#
my $i1, $i2, $s0_last_byte, $s1_last_byte;
$hfe_track_len = ceil( $dsk_track_len * 4 * 2 / 512 ) * 512;

for( $track = 0; $track < $DSK_Tracks*2; $track += 2 ) {
   print "..";
   $i = 0;
   $i1 = 0;
   $i2 = 0;
   while( $i < $hfe_track_len ) {

      # Alternate 512 byte blocks. Pad track with last byte if needed.
      #
      if( $i & 0x0100 ) {
         #
         # Side 1
         #
         if( $i2 < $dsk_track_len*4 ) {
            print OUT chr( $mfm_track[$track+1][$i2] );
            $s1_last_byte = $mfm_track[$track+1][$i2];
            $i2++;
         } else {
            print OUT chr( $s1_last_byte );
         }
      } else {
         #
         # Side 0
         #
         if( $i1 < $dsk_track_len*4 ) {
            print OUT chr( $mfm_track[$track][$i1] );
            $s0_last_byte = $mfm_track[$track][$i1];
            $i1++;
         } else {
            print OUT chr( $s0_last_byte );
         }
      }
   $i++;
   }
}
print "\nDone!\n";

close(OUT);
